#include "GameGenieCPU.h"


unsigned char GameGenieCPU::RAM_Read(Uint16 addr) {
	// First, see if Game Genie wants to mess with read
	for (int i = 0; i < codes.size(); i++) {
		if (codes[i].bln6char) {
			// 6 char code
			if (addr == codes[i].addr) {
				return codes[i].data;
			}
		} else {
			// 8 char code
			if (addr == codes[i].addr) {
				// Replace if memory equals a certain value only
				if ( memory[addr] == codes[i].comp ) {
					return codes[i].data;
				}
			}
		}
	}

	return NesCPU::RAM_Read(addr);
}

void GameGenieCPU::AddCode(char *code) {
	// Decode a 6 or 8 digit code
	strupr( code );

	switch (strlen(code)) {	
	case 6:
		codes.push_back( Decode6CharCode(code) );
		break;
	
	case 8:
		codes.push_back( Decode8CharCode(code) );
		break;
	}
}


CodePatch GameGenieCPU::Decode6CharCode(char *code) {
	CodePatch ret_code;

	// Build 24-bit string (scrambled)
	Uint32 scr_code = 0;
	int shift_amt = 20;
	for (int i = 0; i < 6; i++) {
		Uint8 bit_pattern = CharToBits( code[i] );
		scr_code |= ((bit_pattern & 0x0F) << shift_amt);
		shift_amt -= 4;
	}

	// Unscramble 24-bit string into addr, data
	Uint8 data = 0;
	data |= ((scr_code >> 16) & 0x80);	// bit 24 = bit 1
	data |= ((scr_code >> 12) & 0x40);	// bit 19 = bit 2
	data |= ((scr_code >> 12) & 0x20);	// bit 18 = bit 3
	data |= ((scr_code >> 12) & 0x10);	// bit 17 = bit 4
	data |= ((scr_code >> 0)  & 0x08);	// bit 4  = bit 5
	data |= ((scr_code >> 20) & 0x04);	// bit 23 = bit 6
	data |= ((scr_code >> 20) & 0x02);	// bit 22 = bit 7
	data |= ((scr_code >> 20) & 0x01);  // bit 21 = bit 8
	

	Uint16 addr = 0;
	addr |= ((scr_code << 5)  & 0x8000); // A
	addr |= ((scr_code << 5)  & 0x4000); // B
	addr |= ((scr_code << 5)  & 0x2000); // C
	addr |= ((scr_code << 5)  & 0x1000); // D
	addr |= ((scr_code << 9)  & 0x0800); // E
	addr |= ((scr_code << 9)  & 0x0400); // F
	addr |= ((scr_code << 9)  & 0x0200); // G
	addr |= ((scr_code >> 11) & 0x0100); // H
	addr |= ((scr_code >> 7)  & 0x0080); // I
	addr |= ((scr_code >> 7)  & 0x0040); // J
	addr |= ((scr_code >> 7)  & 0x0020); // K
	addr |= ((scr_code >> 7)  & 0x0010); // L
	addr |= ((scr_code >> 3)  & 0x0008); // M
	addr |= ((scr_code >> 3)  & 0x0004); // N
	addr |= ((scr_code >> 3)  & 0x0002); // O
	
	addr >>= 1;

	addr += 0x8000;

	ret_code.addr = addr;
	ret_code.data = data;
	ret_code.bln6char = true;

	return ret_code;
}



CodePatch GameGenieCPU::Decode8CharCode(char *code) {
	CodePatch ret_code;

	// Build 24-bit string (scrambled)
	Uint32 scr_code = 0;
	int shift_amt = 28;
	for (int i = 0; i < 8; i++) {
		Uint8 bit_pattern = CharToBits( code[i] );
		scr_code |= ((bit_pattern & 0x0F) << shift_amt);
		shift_amt -= 4;
	}

	// Unscramble 32-bit string into addr, data, comp
	Uint8 data = 0;
	data |= ((scr_code >> 24) & 0x80);	// 1
	data |= ((scr_code >> 20) & 0x40);	// 2 
	data |= ((scr_code >> 20) & 0x20);	// 3
	data |= ((scr_code >> 20) & 0x10);	// 4
	data |= ((scr_code >> 0)  & 0x08);	// 5
	data |= ((scr_code >> 28) & 0x04);	// 6
	data |= ((scr_code >> 28) & 0x02);	// 7
	data |= ((scr_code >> 28) & 0x01);  // 8
	

	Uint16 addr = 0;
	addr |= ((scr_code >> 3)  & 0x8000); // A
	addr |= ((scr_code >> 3)  & 0x4000); // B
	addr |= ((scr_code >> 3)  & 0x2000); // C
	addr |= ((scr_code >> 3)  & 0x1000); // D
	addr |= ((scr_code << 1)  & 0x0800); // E
	addr |= ((scr_code << 1)  & 0x0400); // F
	addr |= ((scr_code << 1)  & 0x0200); // G
	addr |= ((scr_code >> 19) & 0x0100); // H
	addr |= ((scr_code >> 15) & 0x0080); // I
	addr |= ((scr_code >> 15) & 0x0040); // J
	addr |= ((scr_code >> 15) & 0x0020); // K
	addr |= ((scr_code >> 15) & 0x0010); // L
	addr |= ((scr_code >> 11) & 0x0008); // M
	addr |= ((scr_code >> 11) & 0x0004); // N
	addr |= ((scr_code >> 11) & 0x0002); // O
	
	addr >>= 1;

	addr += 0x8000;


	Uint8 comp = 0;
	comp |= ((scr_code >> 0) & 0x80);	// !
	comp |= ((scr_code << 4) & 0x40);	// @
	comp |= ((scr_code << 4) & 0x20);	// #
	comp |= ((scr_code << 4) & 0x10);	// $
	comp |= ((scr_code >> 8) & 0x08);	// %
	comp |= ((scr_code >> 4) & 0x04);	// ^
	comp |= ((scr_code >> 4) & 0x02);	// &
	comp |= ((scr_code >> 4) & 0x01);	// *

	ret_code.addr = addr;
	ret_code.data = data;
	ret_code.comp = comp;
	ret_code.bln6char = false;

	return ret_code;
}



Uint8 GameGenieCPU::CharToBits(char c) {
	switch (c) {
	case 'A':
		return 0x0;
		break;
	case 'P':
		return 0x1;
		break;
	case 'Z':
		return 0x2;
		break;
	case 'L':
		return 0x3;
		break;
	case 'G':
		return 0x4;
		break;
	case 'I':
		return 0x5;
		break;
	case 'T':
		return 0x6;
		break;
	case 'Y':
		return 0x7;
		break;
	case 'E':
		return 0x8;
		break;
	case 'O':
		return 0x9;
		break;
	case 'X':
		return 0xA;
		break;
	case 'U':
		return 0xB;
		break;
	case 'K':
		return 0xC;
		break;
	case 'S':
		return 0xD;
		break;
	case 'V':
		return 0xE;
		break;
	case 'N':
		return 0xF;
		break;
	}
	
	return 0;
}








void GameGenieCPU::executeInstruction() {
	// ----------
	// Check for interrupts
	if (mintInterruptLine) {
		if (reg.status & SET_INTERRUPT) {			
			// If the interrupt flag of reg.status is set, then only let NMI interrupts
			// pass through
			if (mintInterruptLine & NMI) {
				mintInterruptLine = NMI;
				handleInterruptRequest();
				return;
			}
		} else {
			handleInterruptRequest();
			return;
		}				
	}
	// -----------

	unsigned int temp, temp2;	// Temporary integers

	//int op = memory[reg.pc];
	int op = RAM_Read(reg.pc);


	// Accumulate clock cycles for timing synchronization.  We add 2 here because at the least,
	// most instructions use 2 instructions.  Later, we add more cycles for various addressing
	// modes and special cases (i.e. page crossings)
	mlngCycleCounter += 2;

	// Calculate the memory index for instruction based on addressing mode into temp2	
	switch( instruction_table[op] ) {
	case IMMEDIATE:
		temp2 = reg.pc + 1;
		reg.pc += 2;
		break;
	case ZERO_PAGE:
		temp2 = RAM_Read(reg.pc+1);
		reg.pc += 2;
		mlngCycleCounter += 1;
		break;
	case ZERO_PAGE_X:				
		temp2 = (RAM_Read(reg.pc+1) + reg.x) & 0xFF;
		reg.pc += 2;
		mlngCycleCounter += 2;
		break;
	case ZERO_PAGE_Y:		
		temp2 = (RAM_Read(reg.pc+1) + reg.y) & 0xFF;
		reg.pc += 2;
		mlngCycleCounter += 2;
		break;
	case INDIRECT:
	case cABSOLUTE:		
		temp2 = RAM_Read(reg.pc+1) + (RAM_Read(reg.pc+2) << 8);
		reg.pc += 3;
		mlngCycleCounter += 2;	
		break;
	case ABSOLUTE_X:
		temp2 = RAM_Read(reg.pc+1) + (RAM_Read(reg.pc+2) << 8) + reg.x;
		reg.pc += 3;
		mlngCycleCounter += 2;
		
		// Add 1 cycle on page crossing except for these instructions:
		switch (op) {
		case ASL_ABSOLUTE_X:
		case DEC_ABSOLUTE_X:
		case INC_ABSOLUTE_X:
		case LSR_ABSOLUTE_X:
		case ROL_ABSOLUTE_X:
		case ROR_ABSOLUTE_X:
			break;
		default:
			if ( ((temp2 - reg.x) & 0xFF00) != (temp2 & 0xFF00) ) mlngCycleCounter++;
			break;
		}
		break;
	case ABSOLUTE_Y:
		temp2 = memory[reg.pc+1] + (memory[reg.pc+2] << 8) + reg.y;
		reg.pc += 3;
		mlngCycleCounter += 2;		
		if ( ((temp2 - reg.y) & 0xFF00) != (temp2 & 0xFF00) ) mlngCycleCounter++;			
		break;

	case INDIRECT_X:		
		temp2 = memory[ (memory[reg.pc+1] + reg.x) & 0xFF ] + (memory[ ((memory[reg.pc+1] + reg.x) & 0xFF) + 1 ] << 8);
		reg.pc += 2;
		mlngCycleCounter += 4;
		break;

	case INDIRECT_Y:
		// Changed 8.29.02				
		//      .... zero page byte 1 ....   ..... zero page byte 2  ............   y reg
		temp2 = memory[ memory[reg.pc+1] ] + (memory[ memory[reg.pc+1] + 1] << 8) + reg.y;
		reg.pc += 2;
		mlngCycleCounter += 3;
		// Add 1 cycle on page crossing except for this instruction
		if (op != STA_INDIRECT_Y) {
			if ( ((temp2 - reg.y) & 0xFF00) != (temp2 & 0xFF00) ) mlngCycleCounter++;			
		}
		break;	
	default:
		break;
	}
		
			
	// Execute Instruction
	switch (op) {
	// Add 1 for page boundary crossings for special modes
	case ADC_INDIRECT_Y:
	case ADC_ABSOLUTE_X:
	case ADC_ABSOLUTE_Y:
		//if ( (reg.pc & 0xFF00) != (temp2 & 0xFF00) ) mlngCycleCounter++;
	case ADC_IMMEDIATE:
	case ADC_ZERO_PAGE:
	case ADC_ZERO_PAGE_X:
	case ADC_ABSOLUTE:	
	case ADC_INDIRECT_X:	
			// Added 9.5.02
			temp = reg.acc + memory[temp2];
			if (reg.status & SET_CARRY) temp++;			
			//reg.status &= CLEAR_ZERO;
			//reg.status &= CLEAR_OVERFLOW;
			//reg.status &= CLEAR_CARRY;
			//reg.status &= CLEAR_SIGN;
			reg.status &= CLEAR_ZOCS;
			if (!(temp & 0xFF))	reg.status |= SET_ZERO;									
			if (temp > 0xFF) reg.status |= SET_CARRY;			
			if (temp & 0x80) reg.status |= SET_SIGN;
			/*if (  (temp > 0xFF) ^ (temp & 0x80) )	// Overflow = XOR of carry and sign
			{										// from nesdev
				reg.status |= SET_OVERFLOW;
			}*/
			if (!((reg.acc ^ memory[temp2]) & 0x80) && ((reg.acc ^ temp) & 0x80)) {
				reg.status |= SET_OVERFLOW;
			}
			reg.acc = (temp & 0xFF);			
			break;
	
	case AND_ABSOLUTE_X:
	case AND_ABSOLUTE_Y:
	case AND_INDIRECT_Y:
		//if ( (reg.pc & 0xFF00) != (temp2 & 0xFF00) ) mlngCycleCounter++;
	case AND_IMMEDIATE:
	case AND_ZERO_PAGE:
	case AND_ZERO_PAGE_X:
	case AND_ABSOLUTE:	
	case AND_INDIRECT_X:	
			reg.acc &= memory[temp2];
			reg.status &= CLEAR_ZERO;
			reg.status &= CLEAR_SIGN;
			if (reg.acc == 0)	reg.status |= SET_ZERO;
			if (reg.acc & 0x80)	reg.status |= SET_SIGN;			
			break;
	
	case ASL_ACCUMULATOR:
			reg.status &= CLEAR_CARRY;
			if (reg.acc & 0x80) reg.status |= SET_CARRY;
			
			// Modified 9.4.02
			reg.acc = (reg.acc << 1) & 0xFF;
			
			reg.status &= CLEAR_SIGN;
			reg.status &= CLEAR_ZERO;			
			if (reg.acc == 0) reg.status |= SET_ZERO;		
			if (reg.acc & 0x80)	reg.status |= SET_SIGN;					
			reg.pc++;			

			break;

	case ASL_ABSOLUTE_X:
			mlngCycleCounter++;
	case ASL_ZERO_PAGE:
	case ASL_ZERO_PAGE_X:
	case ASL_ABSOLUTE:
			mlngCycleCounter += 2;
			temp = memory[temp2];
			RAM_Write(temp2, temp);	// Write-back unmodified value
			reg.status &= CLEAR_CARRY;
			if (temp & 0x80) reg.status |= SET_CARRY;
			
			temp = (temp << 1) & 0xFF;
						
			reg.status &= CLEAR_SIGN;
			reg.status &= CLEAR_ZERO;
			if (temp == 0) reg.status |= SET_ZERO;		
			if (temp & 0x80) reg.status |= SET_SIGN;
			
			//memory[ temp2 ] = temp;
			RAM_Write(temp2, temp);
			break;
		
	case BCC_RELATIVE:			
			temp2 = reg.pc;
			reg.pc += 2;	// PC points to next instruction before displacement			
			if (!(reg.status & SET_CARRY))
			{				
				if (memory[reg.pc-1] & 0x80)
				{					
					reg.pc -= (128 - (memory[reg.pc-1] & 0x7F));
				} else {
					reg.pc += memory[reg.pc-1];
				}
				if ((reg.pc & 0xFF00) == (temp2 & 0xFF00)) {mlngCycleCounter++; } else {mlngCycleCounter += 2;}
			}
			break;			
	
	case BCS_RELATIVE:						
			temp2 = reg.pc;	
			reg.pc += 2;	// PC points to next instruction before displacement
			if ((reg.status & SET_CARRY))
			{
				if (memory[reg.pc-1] & 0x80)
				{
					reg.pc -= (128 - (memory[reg.pc-1] & 0x7F));
				} else {
					reg.pc += memory[reg.pc-1];
				}
				if ((reg.pc & 0xFF00) == (temp2 & 0xFF00)) {mlngCycleCounter++; } else {mlngCycleCounter += 2;}
			} 
			break;			
		
	case BEQ_RELATIVE:						
			temp2 = reg.pc;	
			reg.pc += 2;	// PC points to next instruction before displacement						
			if ((reg.status & SET_ZERO)) {				
				if (memory[reg.pc-1] & 0x80) {
					reg.pc -= (128 - (memory[reg.pc-1] & 0x7F));
				} else {
					reg.pc += memory[reg.pc-1];
				}
				if ((reg.pc & 0xFF00) == (temp2 & 0xFF00)) {mlngCycleCounter++; } else {mlngCycleCounter += 2;}
			} 			
			break;
		
	case BIT_ABSOLUTE:
	case BIT_ZERO_PAGE:			
			temp = RAM_Read(temp2);
			/*reg.status &= CLEAR_SIGN;
			reg.status &= CLEAR_OVERFLOW;
			reg.status &= CLEAR_ZERO;*/
			reg.status &= CLEAR_SOZ;
			if ( (reg.acc & temp) == 0) reg.status |= SET_ZERO;
			if (temp & 0x80) reg.status |= SET_SIGN;
			if (temp & 0x40) reg.status |= SET_OVERFLOW;		
			break;

	case BMI_RELATIVE:			
			temp2 = reg.pc;	
			reg.pc += 2;	// PC points to next instruction before displacement			
			if ((reg.status & SET_SIGN))
			{
				if (memory[reg.pc-1] & 0x80)
				{
					reg.pc -= (128 - (memory[reg.pc-1] & 0x7F));
				}
				else
				{
					reg.pc += memory[reg.pc-1];
				}
				if ((reg.pc & 0xFF00) == (temp2 & 0xFF00)) {mlngCycleCounter++; } else {mlngCycleCounter += 2;}
			} 
			break;

		case BNE_RELATIVE:									
			temp2 = reg.pc;
			reg.pc += 2;	// PC points to next instruction before displacement			
			if (!(reg.status & SET_ZERO)) {
				if (memory[reg.pc-1] & 0x80) {
					reg.pc -= (128 - (memory[reg.pc-1] & 0x7F));
				} else	{
					reg.pc += memory[reg.pc-1];
				}				
				if ((reg.pc & 0xFF00) == (temp2 & 0xFF00)) {mlngCycleCounter++; } else {mlngCycleCounter += 2;}
			}
			break;

		case BPL_RELATIVE:
			temp2 = reg.pc;
			reg.pc += 2;	// PC points to next instruction before displacement			
			if (!(reg.status & SET_SIGN))
			{
				if (memory[reg.pc-1] & 0x80)
				{
					reg.pc -= (128 - (memory[reg.pc-1] & 0x7F));
				}
				else
				{
					reg.pc += memory[reg.pc-1];
				}
				if ((reg.pc & 0xFF00) == (temp2 & 0xFF00)) {mlngCycleCounter++; } else {mlngCycleCounter += 2;}
			} 
			break;

		case BRK_IMPLIED:
			// Trigger a BRK on the interrupt line						
			/*reg.status |= SET_BRK;
			mintInterruptLine |= BRK;
			reg.pc += 2;	// TODO: Some people may argue this ought to be 2*/

			reg.pc += 2;
			memory[reg.sp + 0x100] = ((reg.pc) & 0xFF00) >> 8;	// PCH to reg.sp
			memory[reg.sp - 1 + 0x100] = (reg.pc) & 0xFF;       // PCL to reg.sp-1							
			memory[reg.sp - 2 + 0x100] = reg.status | SET_BRK;		// Status to reg.sp-2, plus BRK flag	
			reg.sp = reg.sp - 3;									// Adjust stack pointer

			reg.status |= SET_INTERRUPT;			
			reg.pc = memory[0xFFFE] + (memory[0xFFFF] << 8);
			mlngCycleCounter += 5;								// BRK takes 7 cycles, need 5 more
			break;

		case BVC_RELATIVE:			
			temp2 = reg.pc;
			reg.pc += 2;	// PC points to next instruction before displacement
			if (!(reg.status & SET_OVERFLOW))
			{
				if (memory[reg.pc-1] & 0x80)
				{
					reg.pc -= (128 - (memory[reg.pc-1] & 0x7F));
				}
				else
				{
					reg.pc += memory[reg.pc-1];
				}
				if ((reg.pc & 0xFF00) == (temp2 & 0xFF00)) {mlngCycleCounter++; } else {mlngCycleCounter += 2;}
			} 
			break;

		case BVS_RELATIVE:
			temp2 = reg.pc;
			reg.pc += 2;	// PC points to next instruction before displacement			
			if ((reg.status & SET_OVERFLOW))
			{
				if (memory[reg.pc-1] & 0x80)
				{
					reg.pc -= (128 - (memory[reg.pc-1] & 0x7F));
				}
				else
				{
					reg.pc += memory[reg.pc-1];
				}
				if ((reg.pc & 0xFF00) == (temp2 & 0xFF00)) {mlngCycleCounter++; } else {mlngCycleCounter += 2;}
			} 
			break;

		case CLC_IMPLIED:			
			reg.status &= CLEAR_CARRY;
			reg.pc++;
			break;
		
		case CLD_IMPLIED:			
			reg.status &= CLEAR_DECIMAL;
			reg.pc++;
			break;
		
		case CLI_IMPLIED:			
			reg.status &= CLEAR_INTERRUPT;
			reg.pc++;
			break;
		
		case CLV_IMPLIED:			
			reg.status &= CLEAR_OVERFLOW;
			reg.pc++;
			break;
		
		case CMP_INDIRECT_Y:
		case CMP_ABSOLUTE_X:
		case CMP_ABSOLUTE_Y:
			//if ( (reg.pc & 0xFF00) != (temp2 & 0xFF00) ) mlngCycleCounter++;		
		case CMP_IMMEDIATE:
		case CMP_ZERO_PAGE:
		case CMP_ZERO_PAGE_X:
		case CMP_ABSOLUTE:		
		case CMP_INDIRECT_X:		
			temp = reg.acc - memory[temp2]; 
			/*reg.status &= CLEAR_SIGN;
			reg.status &= CLEAR_ZERO;
			reg.status &= CLEAR_CARRY;*/
			reg.status &= CLEAR_SZC;
			
			if (temp & 0x80) reg.status |= SET_SIGN;			
			if (!(temp & 0xFF)) reg.status |= SET_ZERO;			
			if (temp < 0x100) reg.status |= SET_CARRY;
			
			break;
		
		case CPX_IMMEDIATE:
		case CPX_ZERO_PAGE:
		case CPX_ABSOLUTE:
			temp = reg.x - memory[temp2];
			/*reg.status &= CLEAR_SIGN;
			reg.status &= CLEAR_ZERO;
			reg.status &= CLEAR_CARRY;*/
			reg.status &= CLEAR_SZC;
			
			if (temp & 0x80) reg.status |= SET_SIGN;			
			if (!(temp & 0xFF)) reg.status |= SET_ZERO;
			if (temp < 0x100) reg.status |= SET_CARRY;
			
			break;
		
		case CPY_IMMEDIATE:
		case CPY_ZERO_PAGE:
		case CPY_ABSOLUTE:
			temp = reg.y - memory[temp2];
			/*reg.status &= CLEAR_SIGN;
			reg.status &= CLEAR_ZERO;
			reg.status &= CLEAR_CARRY;*/
			reg.status &= CLEAR_SZC;
			if (temp & 0x80) reg.status |= SET_SIGN;			
			if (!(temp & 0xFF)) reg.status |= SET_ZERO;
			if (temp < 0x100) reg.status += SET_CARRY;					
			break;
		
		case DEC_ABSOLUTE_X:
			mlngCycleCounter++;		
		case DEC_ZERO_PAGE:
		case DEC_ZERO_PAGE_X:
		case DEC_ABSOLUTE:					
			mlngCycleCounter += 2;
			temp = (memory[ temp2 ] - 1) & 0xFF;
			reg.status &= CLEAR_SIGN;
			reg.status &= CLEAR_ZERO;
			if (temp == 0) reg.status |= SET_ZERO;			
			if (temp & 0x80) reg.status |= SET_SIGN;			
			//memory[ temp2 ] = temp;
			RAM_Write(temp2, temp+1);	// Write-back unmodified value
			RAM_Write(temp2, temp);
			break;

		case DEX_IMPLIED:			
			reg.x = (reg.x - 1) & 0xFF;
			reg.status &= CLEAR_SIGN;
			reg.status &= CLEAR_ZERO;			
			if (reg.x == 0) reg.status += SET_ZERO;			
			if (reg.x & 0x80) reg.status += SET_SIGN;						
			reg.pc++;
			break;

		case DEY_IMPLIED:
			reg.y = (reg.y - 1) & 0xFF;
			reg.status &= CLEAR_SIGN;
			reg.status &= CLEAR_ZERO;			
			if (reg.y == 0) reg.status |= SET_ZERO;			
			if (reg.y & 0x80) reg.status |= SET_SIGN;
			reg.pc++;
			break;
		
		case EOR_INDIRECT_Y:
		case EOR_ABSOLUTE_X:
		case EOR_ABSOLUTE_Y:
			//if ( (reg.pc & 0xFF00) != (temp2 & 0xFF00) ) mlngCycleCounter++;
		case EOR_IMMEDIATE:
		case EOR_ZERO_PAGE:
		case EOR_ZERO_PAGE_X:
		case EOR_ABSOLUTE:		
		case EOR_INDIRECT_X:
		
			reg.acc ^= memory[temp2];
			reg.status &= CLEAR_ZERO;
			reg.status &= CLEAR_SIGN;
			
			if (reg.acc == 0) reg.status |= SET_ZERO;					
			if (reg.acc & 0x80)	reg.status |= SET_SIGN;					
			break;

		case INC_ABSOLUTE_X:
			mlngCycleCounter++;
		case INC_ZERO_PAGE:
		case INC_ZERO_PAGE_X:
		case INC_ABSOLUTE:
			mlngCycleCounter += 2;
			temp = (memory[ temp2 ] + 1) & 0xFF;
			reg.status &= CLEAR_SIGN;
			reg.status &= CLEAR_ZERO;
			
			if (temp == 0) reg.status |= SET_ZERO;			
			if (temp & 0x80) reg.status |= SET_SIGN;
			RAM_Write(temp2, temp-1);	// Write-back unmodified value
			RAM_Write(temp2, temp);
			break;

		case INX_IMPLIED:
			reg.x = (reg.x + 1) & 0xFF;
			reg.status &= CLEAR_SIGN;
			reg.status &= CLEAR_ZERO;
			
			if (!reg.x) reg.status |= SET_ZERO;			
			if (reg.x & 0x80) reg.status |= SET_SIGN;
			
			reg.pc++;
			break;

		case INY_IMPLIED:			
			reg.y = (reg.y + 1) & 0xFF;
			reg.status &= CLEAR_SIGN;
			reg.status &= CLEAR_ZERO;
			
			if (!reg.y) reg.status |= SET_ZERO;			
			if (reg.y & 0x80) reg.status |= SET_SIGN;
			
			reg.pc++;
			break;

		case JMP_ABSOLUTE:			
			mlngCycleCounter++;
			reg.pc = temp2;
			break;

		case JMP_INDIRECT:	
			mlngCycleCounter += 3;
			// CPU bug:  address gets wrapped over page
			reg.pc = memory[temp2] + (memory[temp2 + 1] << 8);
			//reg.pc = memory[temp2] + (memory[ (temp2 & 0xFF00) + ((temp2 + 1) & 0xFF)] << 8);
			break;

		case JSR_ABSOLUTE:			
			memory[reg.sp + 0x100] = ((reg.pc + 2) & 0xFF00) >> 8;	// PCH to reg.sp
			memory[reg.sp - 1 + 0x100] = (reg.pc + 2) & 0xFF;       // PCL to reg.sp-1
			reg.sp -= 2;											// Sub 2 since pc is 16 bits			
			temp = memory[reg.pc + 1] + (memory[reg.pc + 2] << 8);  // New PC is absolute operand
			reg.pc = temp;
			mlngCycleCounter += 2;									// JSR = 6 clock cycles, need 2 more
			break;
	
		case LDA_ABSOLUTE_X:
		case LDA_ABSOLUTE_Y:
		case LDA_INDIRECT_Y:
			//if ( (reg.pc & 0xFF00) != (temp2 & 0xFF00) ) mlngCycleCounter++;
		case LDA_IMMEDIATE:
		case LDA_ZERO_PAGE:
		case LDA_ZERO_PAGE_X:
		case LDA_ABSOLUTE:		
		case LDA_INDIRECT_X:		
			reg.acc = RAM_Read(temp2) & 0xFF;			
			/*reg.status &= CLEAR_ZERO;
			reg.status &= CLEAR_SIGN;*/
			reg.status &= CLEAR_ZS;
			if (reg.acc == 0) reg.status |= SET_ZERO;					
			if (reg.acc & 0x80)	reg.status |= SET_SIGN;								
			break;
		
		
		case LDX_ABSOLUTE_Y:
			//if ( (reg.pc & 0xFF00) != (temp2 & 0xFF00) ) mlngCycleCounter++;
		case LDX_IMMEDIATE:
		case LDX_ZERO_PAGE:
		case LDX_ZERO_PAGE_Y:
		case LDX_ABSOLUTE:					
			reg.x = RAM_Read(temp2) & 0xFF;
			reg.status &= CLEAR_ZERO;
			reg.status &= CLEAR_SIGN;
			if (reg.x == 0)	reg.status |= SET_ZERO;						
			if (reg.x & 0x80) reg.status |= SET_SIGN;							
			break;
		
		case LDY_ABSOLUTE_X:
			//if ( (reg.pc & 0xFF00) != (temp2 & 0xFF00) ) mlngCycleCounter++;
			if ( (temp2 - reg.x) != (temp2 & 0xFF00) ) mlngCycleCounter++;
		case LDY_IMMEDIATE:
		case LDY_ZERO_PAGE:
		case LDY_ZERO_PAGE_X:
		case LDY_ABSOLUTE:					
			reg.y = RAM_Read(temp2) & 0xFF;
			reg.status &= CLEAR_ZERO;
			reg.status &= CLEAR_SIGN;
			
			if (reg.y == 0)	reg.status |= SET_ZERO;					
			if (reg.y & 0x80) reg.status |= SET_SIGN;					
			break;
	
		case LSR_ACCUMULATOR:
			// Added 9.4.02			
			reg.status &= CLEAR_SIGN;
			reg.status &= CLEAR_CARRY;
			reg.status &= CLEAR_ZERO;
			if (reg.acc & 0x1) reg.status |= SET_CARRY;
			
			reg.acc = reg.acc >> 1;
			if (reg.acc == 0) reg.status |= SET_ZERO;
			reg.pc++;
			break;
	
		case LSR_ABSOLUTE_X:
			mlngCycleCounter += 3;	// Need to add 3 more cycles here for 7 total
		case LSR_ZERO_PAGE:			
		case LSR_ZERO_PAGE_X:
		case LSR_ABSOLUTE:		
			// Added 9.4.02			
			reg.status &= CLEAR_SIGN;
			reg.status &= CLEAR_CARRY;
			reg.status &= CLEAR_ZERO;			
			temp = memory[ temp2 ];
			RAM_Write(temp2, temp);	// Write-back unmodified value
			if (temp & 0x1) reg.status |= SET_CARRY;
			
			temp = temp >> 1;
			if (temp == 0) reg.status |= SET_ZERO;									
			//memory[ temp2 ] = temp;						
			RAM_Write(temp2, temp);
			break;
		
		case NOP_IMPLIED:
			reg.pc++;
			break;
		
		case ORA_ABSOLUTE_X:
		case ORA_ABSOLUTE_Y:
		case ORA_INDIRECT_Y:			
			//if ( (reg.pc & 0xFF00) != (temp2 & 0xFF00) ) mlngCycleCounter++;
		case ORA_IMMEDIATE:
		case ORA_ZERO_PAGE:
		case ORA_ZERO_PAGE_X:
		case ORA_ABSOLUTE:		
		case ORA_INDIRECT_X:
			reg.acc |= memory[ temp2 ];
			reg.status &= CLEAR_ZERO;
			reg.status &= CLEAR_SIGN;
			if (reg.acc == 0) reg.status |= SET_ZERO;						
			if (reg.acc & 0x80)	reg.status |= SET_SIGN;						
			break;
		
		case PHA_IMPLIED:			
			memory[reg.sp + 0x100] = reg.acc;
			reg.sp--;
			reg.pc++;
			mlngCycleCounter += 1;					// PHA = 3 cycles, need 1 more
			break;
		
		case PHP_IMPLIED:			
			// BRK flag apparently gets pushed as being set no matter what here
			memory[reg.sp + 0x100] = (reg.status | SET_BRK);
			reg.sp--;
			reg.pc++;
			mlngCycleCounter += 1;
			break;

		case PLA_IMPLIED:			
			reg.sp++;
			reg.acc = memory[reg.sp + 0x100];			
			reg.pc++;
			
			reg.status &= CLEAR_ZERO;
			reg.status &= CLEAR_SIGN;
			if (reg.acc & 0x80) reg.status |= SET_SIGN;
			if (reg.acc == 0) reg.status |= SET_ZERO;
			
			mlngCycleCounter += 2;
			break;
		
		case PLP_IMPLIED:
			reg.sp++;
			reg.status = memory[reg.sp + 0x100] | SET_UNUSED;		
			reg.pc++;
			mlngCycleCounter += 2;
			break;
		
		case ROL_ACCUMULATOR:
			reg.acc = (reg.acc << 1);
			if (reg.status & SET_CARRY) reg.acc |= 0x01;
			reg.status &= CLEAR_CARRY;
			reg.status &= CLEAR_ZERO;
			reg.status &= CLEAR_SIGN;
			if (reg.acc > 0xFF) reg.status |= SET_CARRY;
		
			reg.acc &= 0xFF;
			
			if (reg.acc == 0) reg.status |= SET_ZERO;			
			if (reg.acc & 0x80) reg.status |= SET_SIGN;
	
			reg.pc++;
			break;

		case ROL_ABSOLUTE_X:			
			mlngCycleCounter += 3;	// Need to add 3 more cycles here for 7 total
		case ROL_ZERO_PAGE:
		case ROL_ZERO_PAGE_X:
		case ROL_ABSOLUTE:		
			temp = memory[ temp2 ];
			RAM_Write(temp2, temp);	// Write-back unmodified value
			temp = (temp << 1);
			if (reg.status & SET_CARRY) temp |= 0x01;
			reg.status &= CLEAR_CARRY;
			reg.status &= CLEAR_ZERO;
			reg.status &= CLEAR_SIGN;
			
			if (temp > 0xFF) reg.status |= SET_CARRY;						
			temp &= 0xFF;
			if (temp == 0)  reg.status |= SET_ZERO;			
			if (temp & 0x80) reg.status |= SET_SIGN;
			
			//memory[ temp2 ] = temp;			
			RAM_Write(temp2, temp);
			break;
		
		case ROR_ACCUMULATOR:						
			if (reg.status & SET_CARRY) reg.acc |= 0x100;
			if (reg.acc & 1)
			{
				reg.status |= SET_CARRY;
			} else {
				reg.status &= CLEAR_CARRY;
			}
			
			reg.acc = (reg.acc >> 1);
			
			reg.status &= CLEAR_ZERO;
			reg.status &= CLEAR_SIGN;
																
			if (reg.acc == 0) reg.status |= SET_ZERO;			
			if (reg.acc & 0x80) reg.status |= SET_SIGN;
			
			reg.pc++;
			break;
			
		case ROR_ABSOLUTE_X:
			mlngCycleCounter += 3;	// Need to add 3 more cycles here for 7 total
		case ROR_ZERO_PAGE:
		case ROR_ZERO_PAGE_X:
		case ROR_ABSOLUTE:		
			temp = memory[ temp2 ];
			RAM_Write(temp2, temp);	// Write-back unmodified value
			if (reg.status & SET_CARRY) temp |= 0x100;
			if (temp & 1)
			{
				reg.status |= SET_CARRY;
			} else	{
				reg.status &= CLEAR_CARRY;
			}
			
			temp = (temp >> 1);

			reg.status &= CLEAR_ZERO;
			reg.status &= CLEAR_SIGN;			

			if (temp == 0) reg.status += SET_ZERO;			
			if (temp & 0x80) reg.status += SET_SIGN;
			
			//memory[ temp2 ] = temp;			
			RAM_Write(temp2, temp);
			break;

		case RTI_IMPLIED:			
			reg.sp++;
			reg.status = memory[reg.sp + 0x100];				// Pull off status
			reg.sp++;
			reg.pc = memory[reg.sp + 0x100];					// Pull off PCL
			reg.sp++;
			reg.pc = reg.pc + (memory[reg.sp + 0x100] << 8);	// Pull off PCH
			mlngCycleCounter += 4;								// RTI = 6 cycles, need 4 more			
			break;

		case RTS_IMPLIED:
			reg.sp += 2;
			reg.pc = memory[reg.sp - 1 + 0x100];
			reg.pc = reg.pc + (memory[reg.sp + 0x100] << 8);						
			reg.pc++;
			mlngCycleCounter += 4;						// RTS = 6 cycles, need 4 more
			break;
       
        case SBC_ABSOLUTE_X:
		case SBC_ABSOLUTE_Y:
		case SBC_INDIRECT_Y:
			//if ( (reg.pc & 0xFF00) != (temp2 & 0xFF00) ) mlngCycleCounter++;
		case SBC_IMMEDIATE:
		case SBC_ZERO_PAGE:
		case SBC_ZERO_PAGE_X:
		case SBC_ABSOLUTE:
		case SBC_INDIRECT_X:		
			temp = reg.acc - memory[ temp2 ];
			if (!(reg.status & SET_CARRY)) temp--;
			
			/*reg.status &= CLEAR_ZERO;
			reg.status &= CLEAR_OVERFLOW;
			reg.status &= CLEAR_CARRY;
			reg.status &= CLEAR_SIGN;*/
			
			reg.status &= CLEAR_ZOCS;
			if (!(temp & 0xFF))	reg.status |= SET_ZERO;
								
			/*if (  (temp > 0xFF) ^ (temp & 0x80) )	// Overflow = XOR of carry and sign
			{										// from nesdev
				reg.status |= SET_OVERFLOW;
			}*/
			if (((reg.acc ^ memory[temp2]) & 0x80) && ((reg.acc ^ temp) & 0x80)) {
				reg.status |= SET_OVERFLOW;
			}


			if (temp < 0x100) reg.status |= SET_CARRY;		
			if (temp & 0x80) reg.status |= SET_SIGN;
			
			reg.acc = (temp & 0xFF);			
			break;
	
		case SEC_IMPLIED:
			reg.status &= CLEAR_CARRY;
			reg.status += SET_CARRY;
			reg.pc++;
			break;

		case SED_IMPLIED:
			reg.status &= CLEAR_DECIMAL;
			reg.status += SET_DECIMAL;
			reg.pc++;
			break;

		case SEI_IMPLIED:
			reg.status &= CLEAR_INTERRUPT;
			reg.status += SET_INTERRUPT;
			reg.pc++;
			break;
		
		case STA_ABSOLUTE_X:
		case STA_ABSOLUTE_Y:
		case STA_INDIRECT_Y:
			mlngCycleCounter++;
		case STA_ZERO_PAGE:
		case STA_ZERO_PAGE_X:
		case STA_ABSOLUTE:		
		case STA_INDIRECT_X:
		
			RAM_Write(temp2, reg.acc);
			//memory[ temp2 ] = reg.acc;			
			break;
		
		case STX_ZERO_PAGE:
		case STX_ZERO_PAGE_Y:
		case STX_ABSOLUTE:
			RAM_Write(temp2, reg.x);
			//memory[ temp2 ] = reg.x;			
			break;

		case STY_ZERO_PAGE:
		case STY_ZERO_PAGE_X:
		case STY_ABSOLUTE:
			RAM_Write(temp2, reg.y);
			//memory[ temp2 ] = reg.y;			
			break;
		
		case TAX_IMPLIED:			
			reg.x = reg.acc;
			reg.status &= CLEAR_SIGN;
			reg.status &= CLEAR_ZERO;
			if (reg.x & 0x80) reg.status |= SET_SIGN;
			if (reg.x == 0) reg.status |= SET_ZERO;
			reg.pc++;
			break;

		case TAY_IMPLIED:			
			reg.y = reg.acc;
			reg.status &= CLEAR_SIGN;
			reg.status &= CLEAR_ZERO;
			if (reg.y & 0x80) reg.status |= SET_SIGN;
			if (reg.y == 0) reg.status |= SET_ZERO;
			reg.pc++;
			break;

		case TSX_IMPLIED:			
			// fixed 9.13.02
			reg.x = reg.sp;
			reg.status &= CLEAR_SIGN;
			reg.status &= CLEAR_ZERO;
			if (reg.x & 0x80) reg.status |= SET_SIGN;
			if (reg.x == 0) reg.status |= SET_ZERO;
			reg.pc++;
			break;

		case TXA_IMPLIED:			
			reg.acc = reg.x;
			reg.status &= CLEAR_SIGN;
			reg.status &= CLEAR_ZERO;
			if (reg.acc & 0x80) reg.status |= SET_SIGN;
			if (reg.acc == 0) reg.status |= SET_ZERO;
			reg.pc++;
			break;
		
		case TXS_IMPLIED:			
			reg.sp = reg.x;			
			reg.pc++;
			break;

		case TYA_IMPLIED:		
			reg.acc = reg.y;
			reg.status &= CLEAR_SIGN;
			reg.status &= CLEAR_ZERO;
			if (reg.acc & 0x80) reg.status |= SET_SIGN;
			if (reg.acc == 0) reg.status |= SET_ZERO;
			reg.pc++;
			break;		
		
		default:
			reg.pc++;
	}		

}