//	UberNES Emulation Core
//  Copyright (C) 2004 Luke Molnar
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


//////////////////////////////////////////////////////////////////////
// NesMapper002.cpp - Implementation of mapper #2 (UNROM)
//////////////////////////////////////////////////////////////////////

#include "NesMapper002.h"


NesMapper2::NesMapper2(NesCartridge *linkCartridge, NesCPU *linkCPU, NesPPU *linkPPU) {
	mCartridge = linkCartridge;
	mCpu = linkCPU;
	mPpu = linkPPU;
	
	bytPRGCount = mCartridge->header.PRGROM_count;
	bytCHRCount = mCartridge->header.CHRROM_count;
	
	initMaps();
}


void NesMapper2::initMemory() {	
	if (bytPRGCount > 0) {
		swapPRGBank16(0x8000, 0);			  // First 16k bank to $8000
		swapPRGBank16(0xC000, bytPRGCount-1); // Last 16k bank to $C000
	}
}


void NesMapper2::write(Uint16 addr, unsigned char data) {	
	// Copy a 16K bank of PRG-ROM to $8000	
	swapPRGBank16(0x8000, data % bytPRGCount);	
}