//	UberNES Emulation Core
//  Copyright (C) 2004 Luke Molnar
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


////////////////////////////////////////////////////////////////////
// Nes Mapper #5:  MMC5
// This mapper appears in a few newer NES titles, most notably
// Castlevania 3. Some other games such as Uncharted Waters and
// several Koei titles also use this mapper.
////////////////////////////////////////////////////////////////////

#ifndef NES_MAPPER_5
#define NES_MAPPER_5

#include "../NesBaseMapper.h"


class NesMapper5 : public NesBaseMapper
{
public:
	///////////////////////////////
	// Class Procedures
	///////////////////////////////
	NesMapper5(NesCartridge *linkCartridge, NesCPU *linkCPU, NesPPU *&linkPPU);
	~NesMapper5() { };
	void initMemory();
	void write(Uint16 addr, unsigned char data);
	Uint8 read(Uint16 addr);
	bool blnIrqEnable;
	int intIrqCount, intIrqLatch, intIrqLine, intIrqScanline, intIrqStatus, intIrqClear;

private:
	int intPRGSize, intCHRSize;
	void setCHRROMPages(unsigned char addr, unsigned char data);
	unsigned char CHRMap[8];

};


// We need a special PPU class for scanline counting/IRQ generation
class MMC5PPU : public NesPPU {
	virtual void calcScanLine() {
		// Draw scanline as usual
		NesPPU::calcScanLine();

		// Get a handle to our mapper object, this is pretty convoluted
		NesMapper5 *mmc = (NesMapper5 *) VRAM->mMapper;
	
		if (regs.bytControlReg2 & 0x18) {
			mintCurScanLine -= (mintNumVBlankScanlines + 1);
			if ((mintCurScanLine < 240) && (mintCurScanLine >= 0) ) {
				mmc->intIrqScanline++;
				mmc->intIrqStatus |= 0x40;
				mmc->intIrqClear = 0;
			}
			mintCurScanLine += (mintNumVBlankScanlines + 1);
		}
		if (mmc->intIrqScanline == (mmc->intIrqLine+1)) {
			mmc->intIrqStatus |= 0x80;
		}
		if (++mmc->intIrqClear > 2) {
			mmc->intIrqScanline = 0;
			mmc->intIrqStatus &= ~0xC0;
		}
		if ( (mmc->blnIrqEnable) && (mmc->intIrqStatus & 0x80) && (mmc->intIrqStatus & 0x40) ) {
			mmc->mPpu->timers.blnOtherIRQ = true;
			mmc->mPpu->timers.intOtherPix = 0*3;
			/*mmc->intIrqCount = 0;
			mmc->blnIrqEnable = false;							*/
		}
	}
};

#endif 
