//	UberNES Emulation Core
//  Copyright (C) 2004 Luke Molnar
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


////////////////////////////////////////////////////////////////////
// Nes Mapper #2:  Jaleco SS8806
// This mapper is used on several Japanese titles by Jaleco, such as
// Baseball 3. 
////////////////////////////////////////////////////////////////////

#ifndef NES_MAPPER_18
#define NES_MAPPER_18

#include "../NesBaseMapper.h"


class NesMapper18 : public NesBaseMapper
{
public:
	///////////////////////////////
	// Class Procedures
	///////////////////////////////
	NesMapper18(NesCartridge *linkCartridge, NesCPU *linkCPU, NesPPU *&linkPPU);
	~NesMapper18() { };
	void initMemory();
	void write(Uint16 addr, unsigned char data);

	unsigned char PrgBuf[3];
	unsigned char ChrBuf[8];
	
	int intIrqCount, intIrqLatch;
	bool blnIrqEnable;
};

// We need a special PPU class for scanline counting/IRQ generation
class JalecoPPU : public NesPPU {
	virtual void calcScanLine() {
		// Draw scanline as usual
		NesPPU::calcScanLine();

		// Get a handle to our mapper object, this is pretty convoluted
		NesMapper18 *mmc = (NesMapper18 *) VRAM->mMapper;
	
		// Do IRQ logic
		if (mmc->blnIrqEnable) {
			mmc->intIrqCount -= 113; // apparently IRQ gets set wrt CPU cycles
			if (mmc->intIrqCount < 113) {
				mmc->mPpu->timers.blnOtherIRQ = true;
				mmc->mPpu->timers.intOtherPix = 0*3;
				mmc->intIrqCount = 0;
				mmc->blnIrqEnable = false;
			}				
		}
	}
};


#endif 
