//	UberNES Emulation Core
//  Copyright (C) 2004 Luke Molnar
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

//////////////////////////////////////////////////////////////////////
// NesMapper23.cpp - Implementation of mapper #23 (Konami VRC2 type B)
//////////////////////////////////////////////////////////////////////

#include "NesMapper023.h"



NesMapper23::NesMapper23(NesCartridge *linkCartridge, NesCPU *linkCPU, NesPPU *linkPPU)
{
	mCartridge = linkCartridge;
	mCpu = linkCPU;
	mPpu = linkPPU;
	
	bytPRGCount = mCartridge->header.PRGROM_count;
	bytCHRCount = mCartridge->header.CHRROM_count;
	
	for (int cbanks = 0; cbanks < 8; cbanks++) { chrbanks[cbanks] = 0; }
	initMaps();
	
	intSwitchSpace = 0x8000;
}




// Mapper #21 includes some extra functionality for registers that
// apparently are not used by mapper #23.  So, this routine
// restricts the code of mapper #21 to include only those registers
// that should be included in mapper #23, just in case some
// mapper #23 game tries to write to mapper #21 registers that should
// not have any effect.
//
// Unfortunately, mapper 21 switches around some of the VROM registers
// so a little remapping is needed
//
void NesMapper23::write(Uint16 addr, unsigned char data)
{	
	switch (addr) {
	case 0x8000:
	case 0x9000:
	case 0xA000:
	case 0xB000:
	case 0xB003:
	case 0xC000:
	case 0xC003:
	case 0xD000:
	case 0xD003:
	case 0xE000:
	case 0xE003:
		// These registers are handled exactly the same as in mapper 21
		NesMapper21::write(addr, data);
		break;	
	
	case 0xB001:
	case 0xB002:
	case 0xC001:
	case 0xC002:
	case 0xD001:
	case 0xD002:
	case 0xE001:
	case 0xE002:
		// These registers get switched around (x001 goes to x002 and vice versa)
		NesMapper21::write(addr ^ 0x0003, data);
		break;
	}
}