//	UberNES Emulation Core
//  Copyright (C) 2004 Luke Molnar
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


//////////////////////////////////////////////////////////////////////
// NesMapper071.cpp - Implementation of mapper #71 (Irem 74HC161/32)
//////////////////////////////////////////////////////////////////////

#include "NesMapper078.h"


NesMapper78::NesMapper78(NesCartridge *linkCartridge, NesCPU *linkCPU, NesPPU *linkPPU) {
	mCartridge = linkCartridge;
	mCpu = linkCPU;
	mPpu = linkPPU;
	
	bytPRGCount = mCartridge->header.PRGROM_count;
	bytCHRCount = mCartridge->header.CHRROM_count;
	
	initMaps();

	mPpu->VRAM->mbytMirroringType = SINGLESCREEN;
	mPpu->VRAM->mintSSMirrorLocation = 0x2000;
}


void NesMapper78::initMemory() {	
	if (bytPRGCount > 0) {
		swapPRGBank16(0x8000, 0);			  // First 16k bank to $8000
		swapPRGBank16(0xC000, bytPRGCount-1); // Last 16k bank to $C000
	}
	if (bytCHRCount) swapCHRBank8(0);
}


void NesMapper78::write(Uint16 addr, unsigned char data) {		
	switch (addr & 0x8000) {
	case 0x8000:	// One register: $8000 - $FFFF controls PRGROM and CHRROM
		swapPRGBank16(0x8000, (data & 0x0F) % bytPRGCount );
		swapCHRBank8( ((data & 0xF0) >> 4) % bytCHRCount );

		// Apparently mirroring gets set too
		if ((addr & 0xFE00) != 0xFE00) {
			mPpu->VRAM->mbytMirroringType = SINGLESCREEN;
			mPpu->VRAM->mintSSMirrorLocation = (data & 0x08) ? 0x2400 : 0x2000;
		}
		break;
	}	
}