//	UberNES Emulation Core
//  Copyright (C) 2004 Luke Molnar
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

//////////////////////////////////////////////////////////////////////
// Nes.h: interface for the Nes class.
//////////////////////////////////////////////////////////////////////
// Class Description:  The Nes class is in charge of combining all of
// the NES components (CPU, PPU, Cartridge, etc) into one interface that
// can be used by a client.  The main function this class provides
// is facilitating the loading of a ROM into memory and initializing 
// all of the objects.
//////////////////////////////////////////////////////////////////////

#ifndef NES
#define NES

#include "Porting/basetypes.h"

// Need the following classes for data members:
#include "NesCartridge.h"
#include "NesCPU.h"
#include "NesPPU.h"
#include "NesAPU.h"
#include "NesInput.h"

#include "NesBaseMapper.h"
#include "Mappers/NesMapper001.h"
#include "Mappers/NesMapper002.h"
#include "Mappers/NesMapper003.h"
#include "Mappers/NesMapper004.h"
#include "Mappers/NesMapper005.h"
#include "Mappers/NesMapper006.h"
#include "Mappers/NesMapper007.h"
#include "Mappers/NesMapper009.h"
#include "Mappers/NesMapper010.h"
#include "Mappers/NesMapper011.h"
#include "Mappers/NesMapper016.h"
#include "Mappers/NesMapper018.h"
#include "Mappers/NesMapper019.h"
#include "Mappers/NesMapper021.h"
#include "Mappers/NesMapper022.h"
#include "Mappers/NesMapper023.h"
#include "Mappers/NesMapper024.h"
#include "Mappers/NesMapper025.h"
#include "Mappers/NesMapper032.h"
#include "Mappers/NesMapper033.h"
#include "Mappers/NesMapper034.h"
#include "Mappers/NesMapper064.h"
#include "Mappers/NesMapper065.h"
#include "Mappers/NesMapper066.h"
#include "Mappers/NesMapper069.h"
#include "Mappers/NesMapper071.h"
#include "Mappers/NesMapper075.h"
#include "Mappers/NesMapper078.h"
#include "Mappers/NesMapper079.h"
#include "Mappers/NesMapper228.h"
#include "Mappers/NesMapper232.h"



class Nes  
{
public:
	///////////////////////////////
	// Class Procedures
	///////////////////////////////
	Nes();
	virtual ~Nes();
	virtual int loadCartridge(char *filename);
	virtual int loadCartridge(unsigned char *&cartdata, char *filename);
	virtual void calcFrame();

	virtual void saveState(char *file);
	virtual int loadState(char *file);

	///////////////////////////////
	// Class Data Members
	///////////////////////////////		
	NesCPU *mCpu;									// Pointer to a Nes CPU object
	NesPPU *mPpu;									// Pointer to a Nes PPU object
	NesAPU *mApu;									// POinter to a Nes APU object
	NesCartridge *mCartridge;						// Pointer to a Nes Cartridge object
	NesBaseMapper *mMapper;							// Pointer to a Nes Mapper object
	NesInput *mController1;							// Pointer to a nes input object
	NesInput *mController2;							// Pointer to a nes input object

	bool mblnSRAMUsed;								// Was the SRAM used?
	bool mblnCartLoaded;							// Is a cartridge loaded into the Nes?

private:
	virtual void finishCartridgeLoad(char *filename) { };	

};

#endif 
