//	UberNES Emulation Core
//  Copyright (C) 2004 Luke Molnar
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


//////////////////////////////////////////////////////////////////////
// Nes Input.h: interface for the NesInput class.
//////////////////////////////////////////////////////////////////////
// Class Description:  The NesInput class interfaces devices such as the
// keyboard and joypad with the NES object. 
// The way is does this is slightly tricky but allows for a lot 
// of flexibility:  this class is actually just a "dummy" input class as
// the "resetStrobe" procedure does not poll anything.  If you wanted to 
// interface the keyboard with the NES, you would derive a class from
// this class and implement "resetStrobe" to check the keyboard and fill 
// in the buttonstatus buffer, and then link that new object into the
// NES.  This is complicated, but a demonstration of how to do this
// makes it clear.
//////////////////////////////////////////////////////////////////////


#ifndef NES_INPUT
#define NES_INPUT

#include "Porting/basetypes.h"


#define		NES_A		0x01
#define		NES_B		0x02
#define		NES_SELECT	0x03
#define		NES_START	0x04
#define		NES_UP		0x05
#define		NES_DOWN	0x06
#define		NES_LEFT	0x07
#define		NES_RIGHT	0x08


#define		CONTROLLER1	0x01
#define		CONTROLLER2	0x02

class NesInput
{

public:
	///////////////////////////////
	// Class Procedures
	///////////////////////////////
	NesInput();
	virtual ~NesInput() { 	};	
	NesInput(int intNesDevice);	
	int poll();		
	void clearStrobe();
	virtual void  resetStrobe();

	void saveState(FILE *stream);
	void loadState(FILE *stream);

	///////////////////////////////
	// Class Data Members
	///////////////////////////////
	//bool buttonStatus[9];
	bool buttonStatus[15];
	unsigned char bytStrobeLocation;
	int mintNesDevice;
	int mintErrorCode;	
	unsigned char bytController1Sig; // not used?
	unsigned char bytController2Sig; // not used?
};

#endif 
